# Copyright (c) 2021 SUSE Software Solutions Germany GmbH. All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import warnings
from functools import wraps
from typing import Callable


def obsolete(decommission_at: str, version: str) -> Callable:
    """
    Decorator method to mark an API method as obsolete

    Methods marked with this decorator are still called.
    If the provided decommission date is reached it is
    allowed to decorate the obsoleted method as
    decommissioned.

    Example:

        .. code:: python

            @obsolete(decommission_at='2025-01-28', version='1.2.3')
            def method():
                print('Method implementation...')

    :param str decommission_at: decommision date string.
    :param str version: the first version when this function was deprecated.

    :return: decorated function which is marked as deprecated
    """
    def decorate_obsolete(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f'Function {func.__name__!r} is marked obsolete '
                f'since version {version!r} '
                f'and will be decommissioned at: {decommission_at}'
            )
            return func(*args, **kwargs)
        return wrapper
    return decorate_obsolete


def decommissioned(func: Callable) -> Callable:
    """
    Decorator method to mark an API method as decommissioned

    Methods marked with this decorator are no longer called
    and raises DeprecationWarning when used.
    The method implementation is allowed to be replaced by
    a simple pass statement to get rid of old code

    Example:

        .. code:: python

            @decommissioned
            def method():
                pass

    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        raise DeprecationWarning(
            f'Function {func.__name__!r} has been decommissioned'
        )
    return wrapper



# Copyright (c) 2018 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from abc import (
    ABCMeta,
    abstractmethod
)
from kiwi.oci_tools.base import OCIBase

# project
from kiwi.runtime_config import RuntimeConfig
from kiwi.exceptions import (
    KiwiOCIArchiveToolError
)


class OCI(metaclass=ABCMeta):
    """
    **OCI Factory**
    """
    @abstractmethod
    def __init__(self) -> None:
        return None  # pragma: no cover

    @staticmethod
    def new() -> OCIBase:
        tool_name = RuntimeConfig().get_oci_archive_tool()
        if tool_name == "umoci":
            from kiwi.oci_tools.umoci import OCIUmoci
            return OCIUmoci()
        elif tool_name == "buildah":
            from kiwi.oci_tools.buildah import OCIBuildah
            return OCIBuildah()
        else:
            raise KiwiOCIArchiveToolError(
                'No support for {0} tool available'.format(tool_name)
            )



import io
import os
from unittest.mock import (
    patch, call, Mock, MagicMock
)
from pytest import raises

from kiwi.system.uri import Uri
from kiwi.system.root_import.base import RootImportBase

from kiwi.exceptions import KiwiRootImportError


class TestRootImportBase:
    @patch('os.path.exists')
    @patch('kiwi.system.uri.Defaults.is_buildservice_worker')
    def test_init(self, mock_buildservice, mock_path):
        mock_buildservice.return_value = False
        mock_path.return_value = True
        with patch.dict('os.environ', {'HOME': '../data'}):
            RootImportBase('root_dir', Uri('file:///image.tar.xz'))
        assert call('/image.tar.xz') in mock_path.call_args_list

    def test_init_remote_uri(self):
        with raises(KiwiRootImportError):
            RootImportBase('root_dir', Uri('http://example.com/image.tar.xz'))

    @patch('kiwi.system.root_import.base.log.warning')
    def test_init_unknown_uri(self, mock_log_warn):
        root = RootImportBase('root_dir', Uri('docker://opensuse:leap'))
        assert root.unknown_uri == 'docker://opensuse:leap'
        assert mock_log_warn.called

    @patch('os.path.exists')
    def test_init_non_existing(self, mock_path):
        mock_path.return_value = False
        with patch.dict('os.environ', {'HOME': '../data'}):
            with raises(KiwiRootImportError):
                RootImportBase('root_dir', Uri('file:///image.tar.xz'))

    @patch('os.path.exists')
    def test_data_sync(self, mock_path):
        mock_path.return_value = True
        with patch.dict('os.environ', {'HOME': '../data'}):
            root_import = RootImportBase(
                'root_dir', Uri('file:///image.tar.xz')
            )
        with raises(NotImplementedError):
            root_import.sync_data()

    @patch('os.path.exists')
    def test_overlay_data(self, mock_path):
        mock_path.return_value = True
        with patch.dict('os.environ', {'HOME': '../data'}):
            root_import = RootImportBase(
                'root_dir', Uri('docker://opensuse:leap')
            )
        with raises(NotImplementedError):
            root_import.overlay_data()

    @patch('os.path.exists')
    @patch('kiwi.system.uri.Defaults.is_buildservice_worker')
    @patch('kiwi.system.root_import.base.SystemSetup')
    @patch('kiwi.system.root_import.base.Path')
    @patch('kiwi.system.root_import.base.pathlib')
    @patch('kiwi.system.root_import.base.Command.run')
    @patch('kiwi.system.uri.RuntimeConfig')
    def test_overlay_finalize(
        self, mock_runtime_config, mock_Command_run, mock_pathlib, mock_Path,
        mock_SystemSetup, mock_buildservice, mock_path_exists
    ):
        mock_path_exists.return_value = True
        mock_buildservice.return_value = False
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            xml_state = Mock()
            mock_Command_run.return_value.output = '/file_a\n/file_b'
            with patch.dict('os.environ', {'HOME': '../data'}):
                root = RootImportBase('root_dir', Uri('docker://opensuse:leap'))
                root.overlay = Mock()
                mock_pathlib.Path = Mock()
                root_overlay_path_mock = Mock()
                mock_pathlib.Path.return_value = root_overlay_path_mock

                root.overlay_finalize(xml_state)

                # run config-overlay.sh
                mock_SystemSetup.return_value.\
                    call_config_overlay_script.assert_called_once_with()

                # run config-host-overlay.sh
                mock_SystemSetup.return_value.\
                    call_config_host_overlay_script.assert_called_once_with(
                        working_directory='root_dir'
                    )

                # umount and create the new upper
                root.overlay.umount.assert_called_once_with()
                assert mock_Path.wipe.call_args_list == [
                    call('root_dir'),
                    call('root_dir_cow_before_pinch'),
                    call(root.overlay.lower),
                    call(root.overlay.work)
                ]

                mock_pathlib.Path.assert_called_once_with(root.overlay.upper)
                root_overlay_path_mock.replace.assert_called_once_with('root_dir')

                # find files that got removed
                assert mock_Command_run.call_args_list == [
                    call(
                        [
                            'rsync', '-av', '--dry-run', '--out-format=%n',
                            '--exclude', 'etc/hosts.kiwi',
                            '--exclude', 'etc/hosts.sha',
                            '--exclude', 'etc/resolv.conf.kiwi',
                            '--exclude', 'etc/resolv.conf.sha',
                            '--exclude', 'etc/sysconfig/proxy.kiwi',
                            '--exclude', 'etc/sysconfig/proxy.sha',
                            '--exclude', 'usr/lib/sysimage/rpm',
                            'root_dir_cow_before_pinch/', 'root_dir/'
                        ]
                    ),
                    call(
                        ['find', 'root_dir', '-type', 'c', '-delete']
                    )
                ]
                # create removed files metadata for later host provisioning
                assert file_handle.write.call_args_list == [
                    call('/file_a'), call(os.linesep),
                    call('/file_b'), call(os.linesep)
                ]



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
"""
usage: kiwi-ng -h | --help
       kiwi-ng [--profile=<name>...]
               [--temp-dir=<directory>]
               [--target-arch=<name>]
               [--type=<build_type>]
               [--logfile=<filename>]
               [--logsocket=<socketfile>]
               [--loglevel=<number>]
               [--debug]
               [--debug-run-scripts-in-screen]
               [--color-output]
               [--config=<configfile>]
               [--kiwi-file=<kiwifile>]
           image <command> [<args>...]
       kiwi-ng [--logfile=<filename>]
               [--logsocket=<socketfile>]
               [--loglevel=<number>]
               [--debug]
               [--debug-run-scripts-in-screen]
               [--color-output]
               [--config=<configfile>]
           result <command> [<args>...]
       kiwi-ng [--profile=<name>...]
               [--shared-cache-dir=<directory>]
               [--temp-dir=<directory>]
               [--target-arch=<name>]
               [--type=<build_type>]
               [--logfile=<filename>]
               [--logsocket=<socketfile>]
               [--loglevel=<number>]
               [--debug]
               [--debug-run-scripts-in-screen]
               [--color-output]
               [--config=<configfile>]
               [--kiwi-file=<kiwifile>]
           system <command> [<args>...]
       kiwi-ng -v | --version
       kiwi-ng help

global options:
    --color-output
        use colors for warning and error messages
    --config=<configfile>
        use specified runtime configuration file. If
        not specified the runtime configuration is looked
        up at ~/.config/kiwi/config.yml or /etc/kiwi.yml
    --logfile=<filename>
        create a log file containing all log information including
        debug information even if this was not requested by the
        debug switch. The special call: '--logfile stdout' sends all
        information to standard out instead of writing to a file
    --logsocket=<socketfile>
        send log data to the given Unix Domain socket in the same
        format as with --logfile
    --loglevel=<number>
        specify logging level as number. Details about the
        available log levels can be found at:
        https://docs.python.org/3/library/logging.html#logging-levels
        Setting a log level causes all message >= level to be
        displayed.
    --debug
        print debug information, same as: '--loglevel 10'
    --debug-run-scripts-in-screen
        run scripts called by kiwi in a screen session
    -v --version
        show program version
    help
        show manual page

global options for services: image, system
    --profile=<name>
        profile name, multiple profiles can be selected by passing
        this option multiple times
    --shared-cache-dir=<directory>
        specify an alternative shared cache directory. The directory
        is shared via bind mount between the build host and image
        root system and contains information about package repositories
        and their cache and meta data.
    --temp-dir=<directory>
        specify an alternative base temporary directory. The
        provided path is used as base directory to store temporary
        files and directories. By default /var/tmp is used.
    --type=<build_type>
        image build type. If not set the default XML specified
        build type will be used
    --kiwi-file=<kiwifile>
        Basename of kiwi file which contains the main image
        configuration elements. If not specified kiwi searches for
        a file named config.xml or a file matching *.kiwi

global options for services: image, system
    --target-arch=<name>
        set the image architecture. By default the host architecture is
        used as the image architecture. If the specified architecture name
        does not match the host architecture and is therefore requesting
        a cross architecture image build, it's important to understand that
        for this process to work a preparatory step to support the image
        architecture and binary format on the building host is required
        and not a responsibility of kiwi.
"""
import logging
import sys
import os
from importlib.metadata import entry_points
from docopt import docopt

# project
from kiwi.exceptions import (
    KiwiUnknownServiceName,
    KiwiCommandNotLoaded,
    KiwiLoadCommandUndefined
)
from kiwi.version import __version__
from kiwi.help import Help
from kiwi.defaults import Defaults

log = logging.getLogger('kiwi')


class Cli:
    """
    **Implements the main command line interface**

    An instance of the Cli class builds the entry point for the
    application and implements methods to load further command plugins
    which itself provides their own command line interface
    """
    def __init__(self):
        self.all_args = docopt(
            __doc__,
            version='KIWI (next generation) version ' + __version__,
            options_first=True
        )
        self.command_args = self.all_args['<args>']
        self.command_loaded = None

    def show_and_exit_on_help_request(self):
        """
        Execute man to show the selected manual page
        """
        if self.all_args['help']:
            manual = Help()
            manual.show('kiwi')
            sys.exit(0)

    def get_servicename(self):
        """
        Extract service name from argument parse result

        :return: service name

        :rtype: str
        """
        if self.all_args.get('image') is True:
            return 'image'
        elif self.all_args.get('system') is True:
            return 'system'
        elif self.all_args.get('result') is True:
            return 'result'
        else:
            raise KiwiUnknownServiceName(
                'Unknown/Invalid Servicename'
            )

    def get_command(self):
        """
        Extract selected command name

        :return: command name
        :rtype: str
        """
        return self.all_args['<command>']

    def get_command_args(self):
        """
        Extract argument dict for selected command

        :return:
            Contains dictionary of command arguments

            .. code:: python

                {
                    '--command-option': 'value'
                }

        :rtype: dict
        """
        return self._load_command_args()

    def get_global_args(self):
        """
        Extract argument dict for global arguments

        :return:
            Contains dictionary of global arguments

            .. code:: python

                {
                    '--global-option': 'value'
                }

        :rtype: dict
        """
        result = {}
        for arg, value in list(self.all_args.items()):
            if not arg == '<command>' and not arg == '<args>':
                if arg == '--type' and value == 'vmx':
                    log.warning(
                        'vmx type is now a subset of oem, --type set to oem'
                    )
                    value = 'oem'
                if arg == '--shared-cache-dir' and not value:
                    value = os.sep + Defaults.get_shared_cache_location()
                if arg == '--shared-cache-dir' and value:
                    Defaults.set_shared_cache_location(value)
                if arg == '--temp-dir' and not value:
                    value = Defaults.get_temp_location()
                if arg == '--temp-dir' and value:
                    Defaults.set_temp_location(value)
                if arg == '--target-arch' and value:
                    Defaults.set_platform_name(value)
                if arg == '--config' and value:  # pragma: no cover
                    Defaults.set_custom_runtime_config_file(value)
                result[arg] = value
        return result

    def load_command(self):
        """
        Loads task class plugin according to service and command name

        :return: loaded task module

        :rtype: object
        """
        discovered_tasks = {}
        if sys.version_info >= (3, 12):
            for entry in list(entry_points()):  # pragma: no cover
                if entry.group == 'kiwi.tasks':
                    discovered_tasks[entry.name] = entry.load()
        else:  # pragma: no cover
            module_entries = dict.get(entry_points(), 'kiwi.tasks')
            for entry in module_entries:
                discovered_tasks[entry.name] = entry.load()

        service = self.get_servicename()
        command = self.get_command()

        if not command:
            raise KiwiLoadCommandUndefined(
                'No command specified for {0} service'.format(service)
            )

        self.command_loaded = discovered_tasks.get(
            service + '_' + command
        )
        if not self.command_loaded:
            prefix = 'usage:'
            discovered_tasks_for_service = ''
            for task in discovered_tasks:
                if task.startswith(service):
                    discovered_tasks_for_service += '{0} kiwi-ng {1}\n'.format(
                        prefix, task.replace('_', ' ')
                    )
                    prefix = '      '
            raise KiwiCommandNotLoaded(
                'Command "{0}" not found\n\n{1}'.format(
                    command, discovered_tasks_for_service
                )
            )
        return self.command_loaded

    def _load_command_args(self):
        try:
            argv = [
                self.get_servicename(), self.get_command()
            ] + self.command_args
            return docopt(self.command_loaded.__doc__, argv=argv)
        except Exception:
            raise KiwiCommandNotLoaded(
                '%s command not loaded' % self.get_command()
            )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import importlib
from typing import Dict
from abc import (
    ABCMeta,
    abstractmethod
)

# project
from kiwi.storage.device_provider import DeviceProvider
from kiwi.exceptions import KiwiFileSystemSetupError


class FileSystem(metaclass=ABCMeta):
    """
    **FileSystem factory**

    :param string name: filesystem name
    :param object device_provider: Instance of DeviceProvider
    :param string root_dir: root directory path name
    :param dict custom_args: dict of custom filesystem arguments
    """
    @abstractmethod
    def __init__(self) -> None:
        return None  # pragma: no cover

    @staticmethod
    def new(
        name: str, device_provider: DeviceProvider,
        root_dir: str=None, custom_args: Dict=None  # noqa: E252
    ):
        name_map = {
            'ext2': 'Ext2',
            'ext3': 'Ext3',
            'ext4': 'Ext4',
            'btrfs': 'Btrfs',
            'xfs': 'Xfs',
            'fat16': 'Fat16',
            'fat32': 'Fat32',
            'squashfs': 'SquashFs',
            'swap': 'Swap',
            'erofs': 'EroFs'
        }
        try:
            filesystem = importlib.import_module(
                'kiwi.filesystem.{0}'.format(name)
            )
            return filesystem.__dict__['FileSystem{0}'.format(name_map[name])](
                device_provider, root_dir, custom_args
            )
        except Exception as issue:
            raise KiwiFileSystemSetupError(
                'Support for {0} filesystem not implemented: {1}'.format(
                    name, issue
                )
            )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import logging
from collections import OrderedDict
from typing import (
    Dict, NamedTuple, Tuple
)

# project
from kiwi.defaults import Defaults
from kiwi.utils.temporary import Temporary
from kiwi.command import Command
from kiwi.storage.device_provider import DeviceProvider
from kiwi.storage.mapped_device import MappedDevice
from kiwi.partitioner import Partitioner
from kiwi.runtime_config import RuntimeConfig
from kiwi.exceptions import (
    KiwiCustomPartitionConflictError,
    KiwiError
)

ptable_entry_type = NamedTuple(
    'ptable_entry_type', [
        ('mbsize', int),
        ('clone', int),
        ('partition_name', str),
        ('partition_type', str),
        ('mountpoint', str),
        ('filesystem', str)
    ]
)

log = logging.getLogger('kiwi')


class Disk(DeviceProvider):
    """
    **Implements storage disk and partition table setup**
    """
    def __init__(
        self, table_type: str, storage_provider: DeviceProvider,
        start_sector: int = None, extended_layout: bool = False
    ):
        """
        Construct a new Disk layout object

        :param string table_type: Partition table type name
        :param object storage_provider:
            Instance of class based on DeviceProvider
        :param int start_sector: sector number
        :param bool extended_layout:
            If set to true and on msdos table type when creating
            more than 4 partitions, this will cause the fourth
            partition to be an extended partition and all following
            partitions will be placed as logical partitions inside
            of that extended partition
        """
        self.partition_mapper = RuntimeConfig().get_mapper_tool()
        #: the underlaying device provider
        self.storage_provider = storage_provider

        #: list of protected map ids. If used in a custom partitions
        #: setup this will lead to a raise conditition in order to
        #: avoid conflicts with the existing partition layout and its
        #: customizaton capabilities
        self.protected_map_ids = [
            'root',
            'readonly',
            'boot',
            'prep',
            'spare',
            'swap',
            'efi_csm',
            'efi'
        ]

        #: Unified partition UUIDs according to systemd
        self.gUID = self.get_discoverable_partition_ids()

        self.partition_map: Dict[str, str] = {}
        self.public_partition_id_map: Dict[str, str] = {}
        self.partition_id_map: Dict[str, str] = {}
        self.is_mapped = False

        self.partitioner = Partitioner.new(
            table_type, storage_provider, start_sector, extended_layout
        )

        self.table_type = table_type

    def __enter__(self):
        return self

    def get_device(self) -> Dict[str, MappedDevice]:
        """
        Names of partition devices

        Note that the mapping requires an explicit map() call

        :return: instances of MappedDevice

        :rtype: dict
        """
        device_map = {}
        for partition_name, device_node in list(self.partition_map.items()):
            device_map[partition_name] = MappedDevice(
                device=device_node, device_provider=self
            )
        return device_map

    def is_loop(self) -> bool:
        """
        Check if storage provider is loop based

        The information is taken from the storage provider. If
        the storage provider is loop based the disk is it too

        :return: True or False

        :rtype: bool
        """
        return self.storage_provider.is_loop()

    def create_custom_partitions(
        self, table_entries: Dict[str, ptable_entry_type]
    ) -> None:
        """
        Create partitions from custom data set

        .. code:: python

           table_entries = {
               map_name: ptable_entry_type
           }

        :param dict table: partition table spec
        """
        for map_name in table_entries:
            if map_name in self.protected_map_ids:
                raise KiwiCustomPartitionConflictError(
                    f'Cannot use reserved table entry name: {map_name!r}'
                )
            entry = table_entries[map_name]
            if entry.clone:
                self._create_clones(
                    map_name, entry.clone, entry.partition_type,
                    format(entry.mbsize)
                )
            id_name = f'kiwi_{map_name.title()}Part'
            self.partitioner.create(
                entry.partition_name, entry.mbsize, entry.partition_type
            )
            self._add_to_map(map_name)
            self._add_to_public_id_map(id_name)
            part_uuid = self.gUID.get(entry.partition_name)
            if part_uuid:
                self.partitioner.set_uuid(
                    self.partition_id_map[map_name], part_uuid
                )

    def create_root_partition(self, mbsize: str, clone: int = 0):
        """
        Create root partition

        Populates kiwi_RootPart(id) and kiwi_BootPart(id) if no extra
        boot partition is requested

        :param str mbsize: partition size string
        :param int clone: create [clone] cop(y/ies) of the root partition
        """
        (mbsize, mbsize_clone) = Disk._parse_size(mbsize)
        if clone:
            self._create_clones('root', clone, 't.linux', mbsize_clone)
        self.partitioner.create('p.lxroot', mbsize, 't.linux')
        self._add_to_map('root')
        self._add_to_public_id_map('kiwi_RootPart')
        if 'kiwi_ROPart' in self.public_partition_id_map:
            self._add_to_public_id_map('kiwi_RWPart')
        if 'kiwi_BootPart' not in self.public_partition_id_map:
            self._add_to_public_id_map('kiwi_BootPart')
        root_uuid = self.gUID.get('root')
        if root_uuid:
            self.partitioner.set_uuid(
                self.partition_id_map['root'], root_uuid
            )

    def create_root_lvm_partition(self, mbsize: str, clone: int = 0):
        """
        Create root partition for use with LVM

        Populates kiwi_RootPart(id)

        :param str mbsize: partition size string
        :param int clone: create [clone] cop(y/ies) of the lvm roo partition
        """
        (mbsize, mbsize_clone) = Disk._parse_size(mbsize)
        if clone:
            self._create_clones('root', clone, 't.lvm', mbsize_clone)
        self.partitioner.create('p.lxlvm', mbsize, 't.lvm')
        self._add_to_map('root')
        self._add_to_public_id_map('kiwi_RootPart')
        root_uuid = self.gUID.get('root')
        if root_uuid:
            self.partitioner.set_uuid(
                self.partition_id_map['root'], root_uuid
            )

    def create_root_raid_partition(self, mbsize: str, clone: int = 0):
        """
        Create root partition for use with MD Raid

        Populates kiwi_RootPart(id) and kiwi_RaidPart(id) as well
        as the default raid device node at boot time which is
        configured to be kiwi_RaidDev(/dev/mdX)

        :param str mbsize: partition size string
        :param int clone: create [clone] cop(y/ies) of the raid root partition
        """
        (mbsize, mbsize_clone) = Disk._parse_size(mbsize)
        if clone:
            self._create_clones('root', clone, 't.raid', mbsize_clone)
        self.partitioner.create('p.lxraid', mbsize, 't.raid')
        self._add_to_map('root')
        self._add_to_public_id_map('kiwi_RootPart')
        self._add_to_public_id_map('kiwi_RaidPart')
        root_uuid = self.gUID.get('root')
        if root_uuid:
            self.partitioner.set_uuid(
                self.partition_id_map['root'], root_uuid
            )

    def create_root_readonly_partition(self, mbsize: str, clone: int = 0):
        """
        Create root readonly partition for use with overlayfs

        Populates kiwi_ReadOnlyPart(id), the partition is meant to
        contain a squashfs readonly filesystem. The partition size
        should be the size of the squashfs filesystem in order to
        avoid wasting disk space

        :param str mbsize: partition size string
        :param int clone: create [clone] cop(y/ies) of the ro root partition
        """
        (mbsize, mbsize_clone) = Disk._parse_size(mbsize)
        if clone:
            self._create_clones('root', clone, 't.linux', mbsize_clone)
        self.partitioner.create('p.lxreadonly', mbsize, 't.linux')
        self._add_to_map('readonly')
        self._add_to_public_id_map('kiwi_ROPart')
        root_uuid = self.gUID.get('root')
        if root_uuid:
            self.partitioner.set_uuid(
                self.partition_id_map['readonly'], root_uuid
            )

    def create_boot_partition(self, mbsize: str, clone: int = 0):
        """
        Create boot partition

        Populates kiwi_BootPart(id) and optional kiwi_BootPartClone(id)

        :param str mbsize: partition size string
        :param int clone: create [clone] cop(y/ies) of the boot partition
        """
        (mbsize, mbsize_clone) = Disk._parse_size(mbsize)
        if clone:
            self._create_clones('boot', clone, 't.linux', mbsize_clone)
        self.partitioner.create('p.lxboot', mbsize, 't.linux')
        self._add_to_map('boot')
        self._add_to_public_id_map('kiwi_BootPart')
        boot_uuid = self.gUID.get('xbootldr')
        if boot_uuid:
            self.partitioner.set_uuid(
                self.partition_id_map['boot'], boot_uuid
            )

    def create_prep_partition(self, mbsize: str):
        """
        Create prep partition

        Populates kiwi_PrepPart(id)

        :param str mbsize: partition size string
        """
        (mbsize, _) = Disk._parse_size(mbsize)
        self.partitioner.create('p.prep', mbsize, 't.prep')
        self._add_to_map('prep')
        self._add_to_public_id_map('kiwi_PrepPart')

    def create_spare_partition(self, mbsize: str):
        """
        Create spare partition for custom use

        Populates kiwi_SparePart(id)

        :param str mbsize: partition size string
        """
        (mbsize, _) = Disk._parse_size(mbsize)
        self.partitioner.create('p.spare', mbsize, 't.linux')
        self._add_to_map('spare')
        self._add_to_public_id_map('kiwi_SparePart')

    def create_swap_partition(self, mbsize: str):
        """
        Create swap partition

        Populates kiwi_SwapPart(id)

        :param str mbsize: partition size string
        """
        (mbsize, _) = Disk._parse_size(mbsize)
        self.partitioner.create('p.swap', mbsize, 't.swap')
        self._add_to_map('swap')
        self._add_to_public_id_map('kiwi_SwapPart')
        swap_uuid = self.gUID.get('swap')
        if swap_uuid:
            self.partitioner.set_uuid(
                self.partition_id_map['swap'], swap_uuid
            )

    def create_efi_csm_partition(self, mbsize: str):
        """
        Create EFI bios grub partition

        Populates kiwi_BiosGrub(id)

        :param str mbsize: partition size string
        """
        (mbsize, _) = Disk._parse_size(mbsize)
        self.partitioner.create('p.legacy', mbsize, 't.csm')
        self._add_to_map('efi_csm')
        self._add_to_public_id_map('kiwi_BiosGrub')

    def create_efi_partition(self, mbsize: str):
        """
        Create EFI partition

        Populates kiwi_EfiPart(id)

        :param str mbsize: partition size string
        """
        (mbsize, _) = Disk._parse_size(mbsize)
        self.partitioner.create('p.UEFI', mbsize, 't.efi')
        self._add_to_map('efi')
        self._add_to_public_id_map('kiwi_EfiPart')
        esp_uuid = self.gUID.get('esp')
        if esp_uuid:
            self.partitioner.set_uuid(
                self.partition_id_map['efi'], esp_uuid
            )

    def activate_boot_partition(self):
        """
        Activate boot partition

        Note: not all Partitioner instances supports this
        """
        partition_id = None
        if 'prep' in self.partition_id_map:
            partition_id = self.partition_id_map['prep']
        elif 'boot' in self.partition_id_map:
            partition_id = self.partition_id_map['boot']
        elif 'root' in self.partition_id_map:
            partition_id = self.partition_id_map['root']

        if partition_id:
            self.partitioner.set_flag(partition_id, 'f.active')

    def create_hybrid_mbr(self):
        """
        Turn partition table into a hybrid GPT/MBR table

        Note: only GPT tables supports this
        """
        self.partitioner.set_hybrid_mbr()

    def create_mbr(self):
        """
        Turn partition table into MBR (msdos table)

        Note: only GPT tables supports this
        """
        self.partitioner.set_mbr()

    def set_start_sector(self, start_sector: int):
        """
        Set start sector

        Note: only effective on DOS tables
        """
        self.partitioner.set_start_sector(start_sector)

    def wipe(self):
        """
        Zap (destroy) any GPT and MBR data structures if present
        For DASD disks create a new VTOC table
        """
        if 'dasd' in self.table_type:
            log.debug('Initialize DASD disk with new VTOC table')
            fdasd_input = Temporary().new_file()
            with open(fdasd_input.name, 'w') as vtoc:
                vtoc.write('y\n\nw\nq\n')
            bash_command = ' '.join(
                [
                    'cat', fdasd_input.name, '|',
                    'fdasd', '-f', self.storage_provider.get_device()
                ]
            )
            try:
                Command.run(
                    ['bash', '-c', bash_command]
                )
            except Exception:
                # unfortunately fdasd reports that it can't read in the
                # partition table which I consider a bug in fdasd. However
                # the table was correctly created and therefore we continue.
                # Problem is that we are not able to detect real errors
                # with the fdasd operation at that point.
                log.debug('potential fdasd errors were ignored')
        else:
            log.debug('Initialize %s disk', self.table_type)
            Command.run(
                [
                    'sgdisk', '--zap-all', self.storage_provider.get_device()
                ]
            )

    def map_partitions(self):
        """
        Map/Activate partitions

        In order to access the partitions through a device node it is
        required to map them if the storage provider is loop based
        """
        if self.storage_provider.is_loop():
            if self.partition_mapper == 'kpartx':
                Command.run(
                    ['kpartx', '-s', '-a', self.storage_provider.get_device()]
                )
            else:
                Command.run(
                    ['partx', '--add', self.storage_provider.get_device()]
                )
            self.is_mapped = True
        else:
            Command.run(
                ['partprobe', self.storage_provider.get_device()]
            )

    def get_public_partition_id_map(self) -> Dict[str, str]:
        """
        Populated partition name to number map
        """
        return OrderedDict(
            sorted(self.public_partition_id_map.items())
        )

    def get_discoverable_partition_ids(self) -> Dict[str, str]:
        """
        Ask systemd for a list of standardized GUIDs for the
        current architecture and return them in a dictionary.
        If there is no such information available an empty
        dictionary is returned

        :return: key:value dict from systemd-id128

        :rtype: dict
        """
        discoverable_ids = {}
        try:
            raw_lines = Command.run(
                ['systemd-id128', 'show']
            ).output.split(os.linesep)[1:]
            for line in raw_lines:
                if line:
                    line = ' '.join(line.split())
                    partition_name, uuid = line.split(' ')
                    discoverable_ids[partition_name] = uuid
        except KiwiError as issue:
            log.warning(
                f'Failed to obtain discoverable partition IDs: {issue}'
            )
            log.warning(
                'Using built-in table'
            )
            discoverable_ids = Defaults.get_discoverable_partition_ids()
        return discoverable_ids

    def _create_clones(
        self, name: str, clone: int, type_flag: str, mbsize: str
    ) -> None:
        """
        Create [clone] cop(y/ies) of the given partition name

        The name of a clone partition uses the following name policy:

        * {name}clone{id} for the partition name
        * kiwi_{name}PartClone{id} for the kiwi map name

        :param str name: basename to use for clone partition names
        :param int clone: number of clones, >= 1
        :param str type_flag: partition type name
        :param str mbsize: partition size string
        """
        for clone_id in range(1, clone + 1):
            self.partitioner.create(
                f'p.lx{name}clone{clone_id}', mbsize, type_flag
            )
            self._add_to_map(f'{name}clone{clone_id}')
            self._add_to_public_id_map(f'kiwi_{name}PartClone{clone_id}')

    @staticmethod
    def _parse_size(value: str) -> Tuple[str, str]:
        """
        parse size value. This can be one of the following

        * A number_string
        * The string named: 'all_free'
        * The string formatted as:
              clone:{number_string_origin}:{number_string_clone}

        The method returns a tuple for size and optional clone size
        If no clone size exists both tuple values are the same

        The given number_string for the size of the partition is
        passed along to the actually used partitioner object and
        expected to be valid there. In case invalid size information
        is passed to the partitioner an exception will be raised
        in the scope of the partitioner interface and the selected
        partitioner class

        :param str value: size value

        :return: Tuple of strings

        :rtype: tuple
        """
        if not format(value).startswith('clone:'):
            return (value, value)
        else:
            size_list = value.split(':')
            return (size_list[1], size_list[2])

    def _add_to_public_id_map(self, name, value=None):
        if not value:
            value = self.partitioner.get_id()
        self.public_partition_id_map[name] = value

    def _add_to_map(self, name):
        device_node = None
        partition_number = format(self.partitioner.get_id())
        if self.storage_provider.is_loop():
            device_base = os.path.basename(self.storage_provider.get_device())
            if self.partition_mapper == 'kpartx':
                device_node = ''.join(
                    ['/dev/mapper/', device_base, 'p', partition_number]
                )
            else:
                device_node = ''.join(
                    ['/dev/', device_base, 'p', partition_number]
                )
        else:
            device = self.storage_provider.get_device()
            if device[-1].isdigit():
                device_node = ''.join(
                    [device, 'p', partition_number]
                )
            else:
                device_node = ''.join(
                    [device, partition_number]
                )
        if device_node:
            self.partition_map[name] = device_node
            self.partition_id_map[name] = partition_number

    def __exit__(self, exc_type, exc_value, traceback):
        if self.storage_provider.is_loop() and self.is_mapped:
            log.info('Cleaning up %s instance', type(self).__name__)
            try:
                if self.partition_mapper == 'kpartx':
                    for device_node in self.partition_map.values():
                        Command.run(['dmsetup', 'remove', device_node])
                    Command.run(
                        ['kpartx', '-d', self.storage_provider.get_device()]
                    )
                else:
                    Command.run(
                        ['partx', '--delete', self.storage_provider.get_device()]
                    )
            except Exception as issue:
                log.error(
                    'cleanup of partition maps on {} failed with: {}'.format(
                        self.storage_provider.get_device(), issue
                    )
                )



from unittest.mock import patch

import unittest.mock as mock

from kiwi.defaults import Defaults
from kiwi.filesystem.squashfs import FileSystemSquashFs


class TestFileSystemSquashfs:
    @patch('os.path.exists')
    def setup(self, mock_exists):
        mock_exists.return_value = True
        self.squashfs = FileSystemSquashFs(mock.Mock(), 'root_dir')

    @patch('os.path.exists')
    def setup_method(self, cls, mock_exists):
        self.setup()

    @patch('kiwi.filesystem.squashfs.Command.run')
    def test_create_on_file(self, mock_command):
        Defaults.set_platform_name('x86_64')
        self.squashfs.create_on_file('myimage', 'label')
        mock_command.assert_called_once_with(
            [
                'mksquashfs', 'root_dir', 'myimage', '-noappend',
                '-b', '1M', '-comp', 'xz', '-Xbcj', 'x86'
            ]
        )

    @patch('kiwi.filesystem.squashfs.Command.run')
    def test_create_on_file_exclude_data(self, mock_command):
        Defaults.set_platform_name('ppc64le')
        self.squashfs.create_on_file('myimage', 'label', ['foo'])
        mock_command.assert_called_once_with(
            [
                'mksquashfs', 'root_dir', 'myimage', '-noappend', '-b', '1M',
                '-comp', 'xz', '-Xbcj', 'powerpc', '-wildcards', '-e', 'foo'
            ]
        )

    @patch('kiwi.filesystem.squashfs.Command.run')
    def test_create_on_file_unkown_arch(self, mock_command):
        Defaults.set_platform_name('aarch64')
        self.squashfs.create_on_file('myimage', 'label')
        mock_command.assert_called_once_with(
            [
                'mksquashfs', 'root_dir', 'myimage',
                '-noappend', '-b', '1M', '-comp', 'xz'
            ]
        )

    @patch('kiwi.filesystem.squashfs.Command.run')
    def test_create_on_file_gzip(self, mock_command):
        self.squashfs.custom_args = {
            'compression': 'gzip', 'create_options': []
        }
        self.squashfs.create_on_file('myimage', 'label')
        mock_command.assert_called_once_with(
            [
                'mksquashfs', 'root_dir', 'myimage',
                '-noappend', '-b', '1M', '-comp', 'gzip'
            ]
        )

    @patch('kiwi.filesystem.squashfs.Command.run')
    def test_create_on_file_no_compression(self, mock_command):
        self.squashfs.custom_args = {
            'compression': 'uncompressed', 'create_options': []
        }
        self.squashfs.create_on_file('myimage', 'label')
        mock_command.assert_called_once_with(
            [
                'mksquashfs', 'root_dir', 'myimage', '-noappend',
                '-b', '1M', '-noI', '-noD', '-noF', '-noX'
            ]
        )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import importlib
from abc import (
    ABCMeta,
    abstractmethod
)
import logging

# project
from kiwi.exceptions import KiwiPartitionerSetupError

log = logging.getLogger('kiwi')


class Partitioner(metaclass=ABCMeta):
    """
    **Partitioner factory**

    :param string table_type: Table type name
    :param object storage_provider: Instance of class based on DeviceProvider
    :param int start_sector: sector number
    :param bool extended_layout: support extended layout for msdos table
    """
    @abstractmethod
    def __init__(self) -> None:
        return None  # pragma: no cover

    @staticmethod
    def new(
        table_type: str, storage_provider: object,
        start_sector: int = None, extended_layout: bool = False
    ):
        name_map = {
            'msdos': 'MsDos',
            'gpt': 'Gpt',
            'dasd': 'Dasd'
        }
        try:
            partitioner = importlib.import_module(
                'kiwi.partitioner.{0}'.format(table_type)
            )
            module_name = 'Partitioner{0}'.format(name_map[table_type])
            if table_type == 'dasd' and start_sector:
                log.warning(
                    'disk start_sector value is ignored for dasd partitions'
                )
                start_sector = None
            return partitioner.__dict__[module_name](
                storage_provider, start_sector, extended_layout
            )
        except Exception as issue:
            raise KiwiPartitionerSetupError(
                'Support for {0} partitioner not implemented: {1}'.format(
                    table_type, issue
                )
            )



from unittest.mock import patch, call
from pytest import (
    raises, fixture
)
import unittest.mock as mock

from kiwi.package_manager.pacman import PackageManagerPacman

from kiwi.exceptions import KiwiRequestError


class TestPackageManagerPacman:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def setup(self):
        repository = mock.Mock()
        repository.root_dir = '/root-dir'

        repository.runtime_config = mock.Mock(
            return_value={
                'pacman_args': [
                    '--config', '/root-dir/pacman.conf', '-y', '--noconfirm'
                ],
                'command_env': ['env']
            }
        )
        self.manager = PackageManagerPacman(repository)

    def setup_method(self, cls):
        self.setup()

    def test_request_package(self):
        self.manager.request_package('name')
        assert self.manager.package_requests == ['name']

    def test_request_collection(self):
        self.manager.request_collection('name')
        assert self.manager.collection_requests == []
        assert self.manager.package_requests == ['name']

    def test_request_product(self):
        self.manager.request_product('name')
        assert self.manager.product_requests == []

    def test_request_package_exclusion(self):
        self.manager.request_package_exclusion('name')
        assert self.manager.exclude_requests == ['name']

    def test_setup_repository_modules(self):
        self.manager.setup_repository_modules({})

    @patch('kiwi.command.Command.call')
    @patch('kiwi.command.Command.run')
    def test_process_install_requests_bootstrap(self, mock_run, mock_call):
        self.manager.request_package('vim')
        self.manager.request_collection('collection')
        self.manager.process_install_requests_bootstrap()
        mock_run.call_args_list == [
            call(['rm', '-r', '-f', '/root-dir/var/run']),
            call([
                'pacman', '--config', '/root-dir/pacman.conf', '-y',
                '--noconfirm', '--needed', '--root', '/root-dir', '-Sy'
            ])
        ]
        mock_call.assert_called_once_with(
            [
                'pacman', '--config', '/root-dir/pacman.conf',
                '-y', '--noconfirm', '--root', '/root-dir', '-S',
                '--needed', '--overwrite', '/root-dir/var/run',
                'vim', 'collection'
            ], ['env']
        )

    @patch('kiwi.command.Command.call')
    def test_process_install_requests(self, mock_call):
        self.manager.request_package('vim')
        self.manager.request_collection('collection')
        self.manager.request_package_exclusion('skipme')
        self.manager.process_install_requests()
        mock_call.assert_called_once_with(
            [
                'chroot', '/root-dir', 'pacman', '--config',
                '/pacman.conf', '-y', '--noconfirm', '-S',
                '--needed', 'vim', 'collection'
            ], ['env']
        )

    @patch('kiwi.command.Command.call')
    @patch('kiwi.command.Command.run')
    def test_process_delete_requests_force(self, mock_run, mock_call):
        self.manager.request_package('vim')
        self.manager.process_delete_requests(True)
        mock_call.assert_called_once_with(
            [
                'chroot', '/root-dir', 'pacman', '--config', '/pacman.conf',
                '-y', '--noconfirm', '-Rdd', 'vim'
            ], ['env']
        )

    @patch('kiwi.command.Command.call')
    @patch('kiwi.command.Command.run')
    def test_process_delete_requests_no_force(self, mock_run, mock_call):
        self.manager.request_package('vim')
        self.manager.process_delete_requests()
        mock_call.assert_called_once_with(
            [
                'chroot', '/root-dir', 'pacman', '--config', '/pacman.conf',
                '-y', '--noconfirm', '-Rs', 'vim'
            ], ['env']
        )

    @patch('kiwi.command.Command.run')
    @patch('kiwi.command.Command.call')
    def test_process_delete_requests_package_missing(
        self, mock_call, mock_run
    ):
        mock_run.side_effect = Exception
        self.manager.request_package('vim')
        with raises(KiwiRequestError):
            self.manager.process_delete_requests()
        mock_run.assert_called_once_with(
            ['chroot', '/root-dir', 'pacman', '-Qi', 'vim']
        )

    def test_process_only_required(self):
        self.manager.process_only_required()
        assert self.manager.custom_args == []

    def test_process_plus_recommended(self):
        self.manager.process_plus_recommended()
        assert self.manager.custom_args == []

    @patch('kiwi.command.Command.call')
    def test_update(self, mock_call):
        self.manager.update()
        mock_call.assert_called_once_with(
            [
                'chroot', '/root-dir', 'pacman', '--config', '/pacman.conf',
                '-y', '--noconfirm', '-Su'
            ], ['env']
        )

    def test_match_package_installed(self):
        assert self.manager.match_package_installed('foo', ' installing foo')

    def test_match_package_deleted(self):
        assert self.manager.match_package_deleted('foo', ' removing foo')



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import re
import logging
from typing import (
    List, Dict
)


# project
from kiwi.command import CommandCallT
from kiwi.command import Command
from kiwi.utils.rpm_database import RpmDataBase
from kiwi.utils.rpm import Rpm
from kiwi.package_manager.base import PackageManagerBase
from kiwi.system.root_bind import RootBind
from kiwi.path import Path
from kiwi.defaults import Defaults
from kiwi.repository.dnf4 import RepositoryDnf4

from kiwi.exceptions import KiwiRequestError

log = logging.getLogger('kiwi')


class PackageManagerMicroDnf(PackageManagerBase):
    """
    **Implements Installation/Deletion of packages/collections with microdnf***

    :param doct dnf_args:
        microdnf arguments from repository runtime configuration
    :param dict command_env:
        microdnf command environment from repository runtime
        configuration
    """
    def post_init(self, custom_args: List = []) -> None:
        """
        Post initialization method

        :param list custom_args: custom microdnf arguments
        """
        self.repository: RepositoryDnf4 = self.repository
        self.custom_args = custom_args

        runtime_config = self.repository.runtime_config()
        self.dnf_args = runtime_config['dnf_args']
        self.command_env = runtime_config['command_env']

    def request_package(self, name: str) -> None:
        """
        Queue a package request

        :param str name: package name
        """
        self.package_requests.append(name)

    def request_collection(self, name: str) -> None:
        """
        Queue a collection request

        :param str name: dnf group name
        """
        log.warning(
            'Group(%s) handling not yet supported for microdnf', name
        )

    def request_product(self, name: str) -> None:
        """
        Queue a product request

        There is no product definition in the fedora repo data

        :param str name: unused
        """
        pass

    def request_package_exclusion(self, name: str) -> None:
        """
        Queue a package exclusion(skip) request

        :param str name: package name
        """
        self.exclude_requests.append(name)

    def setup_repository_modules(
        self, collection_modules: Dict[str, List[str]]
    ) -> None:
        """
        Setup repository modules and streams

        :param dict collection_modules:
            Expect dict of the form:

            .. code:: python

                {
                    'enable': [
                        "module:stream", "module"
                    ],
                    'disable': [
                        "module"
                    ]
                }
        """
        microdnf_module_command = [
            'microdnf'
        ] + ['--refresh'] + self.dnf_args + [
            '--installroot', self.root_dir,
            f'--releasever={self.release_version}',
            '--noplugins',
            '--setopt=cachedir={0}'.format(
                self.repository.shared_dnf_dir['cache-dir']
            ),
            '--setopt=reposdir={0}'.format(
                self.repository.shared_dnf_dir['reposd-dir']
            ),
            '--setopt=varsdir={0}'.format(
                self.repository.shared_dnf_dir['vars-dir']
            )
        ] + self.custom_args + [
            'module'
        ]
        for disable_module in collection_modules['disable']:
            Command.run(
                microdnf_module_command + [
                    'disable', disable_module
                ], self.command_env
            )
        for enable_module in collection_modules['enable']:
            Command.run(
                microdnf_module_command + [
                    'reset', enable_module.split(':')[0]
                ], self.command_env
            )
            Command.run(
                microdnf_module_command + [
                    'enable', enable_module
                ], self.command_env
            )

    def process_install_requests_bootstrap(
        self, root_bind: RootBind = None, bootstrap_package: str = None
    ) -> CommandCallT:
        """
        Process package install requests for bootstrap phase (no chroot)

        :param object root_bind: unused
        :param str bootstrap_package: unused

        :return: process results in command type

        :rtype: namedtuple
        """
        microdnf_command = [
            'microdnf'
        ] + ['--refresh'] + self.dnf_args + [
            '--installroot', self.root_dir,
            f'--releasever={self.release_version}',
            '--noplugins',
            '--setopt=cachedir={0}'.format(
                self.repository.shared_dnf_dir['cache-dir']
            ),
            '--setopt=reposdir={0}'.format(
                self.repository.shared_dnf_dir['reposd-dir']
            ),
            '--setopt=varsdir={0}'.format(
                self.repository.shared_dnf_dir['vars-dir']
            )
        ] + self.custom_args + ['install'] + self.package_requests
        self.cleanup_requests()
        return Command.call(
            microdnf_command, self.command_env
        )

    def process_install_requests(self) -> CommandCallT:
        """
        Process package install requests for image phase (chroot)

        :return: process results in command type

        :rtype: namedtuple
        """
        exclude_args = []
        if self.exclude_requests:
            # For DNF, excluding a package means removing it from
            # the solver operation. This is done by adding --exclude
            # to the command line. This means that if the package is
            # hard required by another package, it will break the transaction.
            for package in self.exclude_requests:
                exclude_args.append('--exclude=' + package)
        chroot_dnf_args = Path.move_to_root(
            self.root_dir, self.dnf_args
        )
        microdnf_command = [
            'chroot', self.root_dir, 'microdnf'
        ] + chroot_dnf_args + [
            f'--releasever={self.release_version}'
        ] + self.custom_args + exclude_args + [
            'install'
        ] + self.package_requests
        self.cleanup_requests()
        return Command.call(
            microdnf_command, self.command_env
        )

    def process_delete_requests(self, force: bool = False) -> CommandCallT:
        """
        Process package delete requests (chroot)

        :param bool force: force deletion: true|false

        :raises KiwiRequestError: if none of the packages to delete is
            installed.
        :return: process results in command type

        :rtype: namedtuple
        """
        if force:
            delete_items = []
            for delete_item in self.package_requests:
                try:
                    Command.run(
                        ['chroot', self.root_dir, 'rpm', '-q', delete_item]
                    )
                    delete_items.append(delete_item)
                except Exception:
                    # ignore packages which are not installed
                    pass
            if not delete_items:
                raise KiwiRequestError(
                    'None of the requested packages to delete are installed'
                )
            self.cleanup_requests()
            delete_options = ['--nodeps', '--allmatches', '--noscripts']
            return Command.call(
                [
                    'chroot', self.root_dir, 'rpm', '-e'
                ] + delete_options + delete_items,
                self.command_env
            )
        else:
            chroot_dnf_args = Path.move_to_root(self.root_dir, self.dnf_args)
            dnf_command = [
                'chroot', self.root_dir, 'microdnf'
            ] + chroot_dnf_args + [
                f'--releasever={self.release_version}'
            ] + self.custom_args + [
                'remove'
            ] + self.package_requests
            self.cleanup_requests()
            return Command.call(
                dnf_command, self.command_env
            )

    def update(self) -> CommandCallT:
        """
        Process package update requests (chroot)

        :return: process results in command type

        :rtype: namedtuple
        """
        chroot_dnf_args = Path.move_to_root(self.root_dir, self.dnf_args)
        return Command.call(
            [
                'chroot', self.root_dir, 'microdnf'
            ] + chroot_dnf_args + [
                f'--releasever={self.release_version}'
            ] + self.custom_args + [
                'upgrade'
            ],
            self.command_env
        )

    def process_only_required(self) -> None:
        """
        Setup package processing only for required packages
        """
        if '--setopt=install_weak_deps=0' not in self.custom_args:
            self.custom_args.append('--setopt=install_weak_deps=0')

    def process_plus_recommended(self) -> None:
        """
        Setup package processing to also include recommended dependencies.
        """
        if '--setopt=install_weak_deps=0' in self.custom_args:
            self.custom_args.remove('--setopt=install_weak_deps=0')

    def match_package_installed(
        self, package_name: str, package_manager_output: str
    ) -> bool:
        """
        Match expression to indicate a package has been installed

        This match for the package to be installed in the output
        of the microdnf command is not 100% accurate. There might
        be false positives due to sub package names starting with
        the same base package name

        :param str package_name: package_name
        :param str package_manager_output: microdnf status line

        :returns: True|False

        :rtype: bool
        """
        return bool(
            re.match(
                '.*Installing  : {0}.*'.format(re.escape(package_name)),
                package_manager_output
            )
        )

    def match_package_deleted(
        self, package_name: str, package_manager_output: str
    ) -> bool:
        """
        Match expression to indicate a package has been deleted

        :param str package_name: package_name
        :param str package_manager_output: microdnf status line

        :returns: True|False

        :rtype: bool
        """
        return bool(
            re.match(
                '.*Removing: {0}.*'.format(re.escape(package_name)),
                package_manager_output
            )
        )

    def post_process_install_requests_bootstrap(
        self, root_bind: RootBind = None, delta_root: bool = False
    ) -> None:
        """
        Move the rpm database to the place as it is expected by the
        rpm package installed during bootstrap phase

        :param object root_bind: unused
        :param bool delta_root: unused
        """
        rpmdb = RpmDataBase(self.root_dir)
        if rpmdb.has_rpm():
            rpmdb.rebuild_database()
            rpmdb.set_database_to_image_path()

    def clean_leftovers(self) -> None:
        """
        Cleans package manager related data not needed in the
        resulting image such as custom macros
        """
        Rpm(
            self.root_dir, Defaults.get_custom_rpm_image_macro_name()
        ).wipe_config()
